class OutputDataset:
    """Marker type for output datasets."""


class InputDataset:
    """Marker type for input datasets."""


class InputModel:
    """Marker type for input models."""


class OutputModel:
    """Marker type for output models."""
